! ========================================================================
! Copyright (C) by GemTalk Systems 1991-2016.  All Rights Reserved
!
! $Id: multibytestring.gs,v 1.14.6.1 2008-04-16 18:43:00 dhenrich Exp $
!
! ========================================================================

! Create MultiByteSymbol class if not already defined (as during conversion),

expectvalue %String
run
| res cls oldCls clsName hst |
clsName := #MultiByteString.
oldCls := Globals at:clsName otherwise: nil .
oldCls == nil ifTrue:[

  " Must be converting a pre 2.3 DB "
  " Create MultiByteString "
  res := CharacterCollection _newKernelSubclass: clsName
    instVarNames: #()
    classVars: #()
    classInstVars: #()
    poolDictionaries: #[]
    inDictionary: Globals
    constraints: #[] 
    instancesInvariant: false
    isModifiable: true
    reservedOop: 989.
    
  cls := Globals at: clsName.

  "_newKernelSubclass creates oop objects: change format to byte/indexable"
  cls _unsafeAt: 2 put: 5.

  cls immediateInvariant.

  "Modify DoubleByteString to have MultiByteString as its superClass "
  DoubleByteString _unsafeAt: 1 put: cls.
  DoubleByteString class _unsafeAt: 1 put: cls class.

  "Replace DoubleByteString with MultiByteString in the classHistory
   of String "
  hst := String classHistory.
  hst remove: DoubleByteString ifAbsent: [].
  hst addLast: MultiByteString.

] ifFalse: [
  "MultiByteString class exists"
  res := clsName asString , ' already exists'.

  "Check if DoubleByteString needs to have metaclass fix (bug 41637)"
  (DoubleByteString class superClass == oldCls class) ifFalse: [
     DoubleByteString class _unsafeAt: 1 put: oldCls class.
     res addAll: ', but needed metaclass updated to MultiByteString class'.
  ]].

^ res
%

removeallmethods MultiByteString
removeallclassmethods MultiByteString

category: 'For Documentation Installation only'
classmethod: MultiByteString
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'MultiByteString is an abstract class for representing strings
 for which each Character occupies more than one byte.

 MultiByteString is in the ClassHistory of String, so instances of the
 subclasses of MultiByteString may be stored into instance variables that 
 are constrained to hold instances of String.  The inverse is not true, 
 so in an application that uses a mixture of MultiByteStrings and Strings,
 string constraints should always be expressed as String.' .
doc documentClassWith: txt.

txt := (GsDocText new) details:
'Unless noted otherwise, comparisons between MultiByteStrings are 
 case-sensitive.

 Some of these methods determine whether one strings collates before
 another.  In collation, the Characters of the receiver and aString are
 compared Character-by-Character, from left to right, in case-sensitive
 fashion.  If two strings are of different length, and all Characters in the
 shorter string are equal to their counterparts in the longer strings, the
 shorter string collates before the longer.

 These methods require that the argument be a kind of CharacterCollection, not
 necessarily a String.'.
doc documentCategory: #Comparing with: txt.

self description: doc.
%

! ------------------- Class methods for MultiByteString

category 'Formatting'
classmethod: MultiByteString
charSize

"Returns number of bytes that make up a character for this string class."

^ self subclassResponsibility: #charSize
%

category 'Instance Creation'
classmethod: MultiByteString
new

^ self subclassResponsibility: #new
%

category 'Instance Creation'
classmethod: MultiByteString
new: aSize

^ self subclassResponsibility: #new:
%

category 'Instance Creation'
classmethod: MultiByteString
withAll: aString

"Returns an instance of the receiver containing the elements of the argument."

<primitive: 456>

^ super withAll: aString
%

! ------------------- Instance methods for MultiByteString
category 'Formatting'
method: MultiByteString
charSize

"Returns number of bytes that make up a character for this string class."

^ self subclassResponsibility: #charSize
%

category: 'Formatting'
method: MultiByteString
printOn: aStream

"Puts a displayable representation of the receiver on the given stream."

"For classes whose instances can be literals, the result should contain
 formatting information.  For example, the following expression should
 evaluate to true:

 #abc asString = String withAll: '#abc'"

| subStr idx lastIdx sz |

aStream _nextPut: $' .

idx := self indexOf: $' startingAt: 1 .
(idx == 0 _or:[ idx > aStream maxSize ]) ifTrue:[
  aStream nextPutAll: self .
  ]
ifFalse:[
  subStr := String new .
  lastIdx := 1.
  [ (idx == 0) _or: [aStream isFull]] whileFalse: [
    self copyFrom: lastIdx to: idx into: subStr startingAt: 1 .
    aStream nextPutAll: subStr .
    subStr size: 0 .
    aStream nextPut: $' .
    lastIdx := idx + 1 .
    idx := self indexOf: $' startingAt: lastIdx .
    ].
  sz := self size .
  ((lastIdx <= sz) _and: [aStream isFull not]) ifTrue: [
    self copyFrom: lastIdx to: sz into: subStr startingAt: 1 .
    aStream nextPutAll: subStr .
    subStr size: 0 .
    ].
  ].
aStream _nextPut: $' 
%

category: 'Concatenating'
method: MultiByteString
, aCharOrCharColl

"Returns a new instance of the receiver's class that contains the elements of
 the receiver followed by the elements of aCharOrCharColl.  The argument
 must be a String, a MultiByteString, or an AbstractCharacter.

 The result may not be an instance of the class of the receiver if one of the
 following rules applies:

 1) If the receiver or argument is a kind of Symbol, MultiByteSymbol, or 
    ObsoleteInvariantString, the result is a MultiByteString.

 2) If the argument is not a MultiByteString, and is a subclass of
    MultiByteString, then the result is an instance of the class of the
    argument.

 Warning: Creating a new instance and copying the receiver take time.  If you
 can safely modify the receiver, it can be much faster to use the addAll:
 method.  See the documentation of the Concatenating category of class
 SequenceableCollection for more details."

<primitive: 614 >

| result |
(aCharOrCharColl isKindOf: String) ifTrue:[
  result := self class new.
  result addAll: self ; addAll: aCharOrCharColl .
  ^ result
  ].

aCharOrCharColl _validateClasses: #[ AbstractCharacter, CharacterCollection].
^ super , aCharOrCharColl
%

! bugfix 39431: check when to return original string

category: 'Converting'
method: MultiByteString
asString

"Returns a String representation of the receiver."

| str max |
(self charSize == 2) 
  ifTrue:  [ max := 255 ]    "For DoubleByteStrings"
  ifFalse: [ max := 65535 ]. "For QuadByteStrings"
str := String new.
self do: [:c | 
   (c asInteger > max) ifTrue: [ ^ self ].
   str add: c].
^ str
%

category: 'Converting'
method: MultiByteString
asByteArray

	^ByteArray fromString: self
%

category: 'Converting'
method: MultiByteString
asDoubleByteSymbol

"Returns a symbol representation of the receiver."

^ DoubleByteSymbol withAll: self.
%

category: 'Converting'
method: MultiByteString
_asString

"Returns a (single byte) String representation of the receiver if all of the
 Characters in the receiver can be represented as single-byte Characters.
 Returns nil if any of the Characters in the receiver cannot be represented as
 single-byte Characters."

<primitive: 229> 

^ self _primitiveFailed: #_asString .
%

category: 'Converting'
method: MultiByteString
asSymbol

"Returns a canonical Symbol representation of the receiver."

^ Symbol withAll: self .
%

category: 'Converting'
method: MultiByteString
asSymbolKind

"Returns a (single byte) Symbol representation of the receiver."

^ Symbol withAll: self 
%

! Fix 36125
category: 'Accessing'
method: MultiByteString
wordAt: anIndex

"Returns the integer value of the Character at anIndex in the receiver."

^ (self at: anIndex) asciiValue
%

method:
codePointAt: anIndex

"Returns the integer value of the Character at anIndex in the receiver."

^ (self at: anIndex) asciiValue
%

category: 'Updating'
method: MultiByteString
wordAt: anIndex put: aValue

"Stores the integer value aValue in the character cell of the receiver
 specified by anIndex. Return aValue."

self at: anIndex put: (Character withValue: aValue) .
^ aValue
%

category: 'Accessing'
method: MultiByteString
at: anIndex

"Returns the Character at anIndex."

^ self subclassResponsibility: #at:
%

category: 'Updating'
method: MultiByteString
at: anIndex put: aChar

"(R) Stores aChar at the specified location and returns aChar."

<primitive: 436>

(aChar _class == Character) ifTrue:[ 
    aChar asciiValue > 16rFFFF ifTrue:[
      ^ self _convertToQuadByte at:anIndex put:aChar ].
  ] ifFalse:[
    aChar _validateClass: AbstractCharacter .
    ^ self at: anIndex put: aChar asCharacter ].
(anIndex _isInteger) ifTrue: [
  ((anIndex > (self size + 1)) _or: [anIndex <= 0]) ifTrue: [
    ^ self _errorIndexOutOfRange: anIndex
  ].
] ifFalse:[ 
  ^ self _errorNonIntegerIndex: anIndex
] .
self _primitiveFailed: #at:put:
%

category: 'Case-Sensitive Comparisons'
method: MultiByteString
= aCharCollection

"Returns true if corresponding Characters in the receiver and argument are equal
 and aCharCollection is comparable with the receiver, and aCharCollection is
 not a kind of Symbol.  Returns false otherwise.

 The comparison for 'equal' is case-sensitive .

 Note that 'kind of Symbol' means either an instance of Symbol or instance of
 MultiByteSymbol."

<primitive: 449>

(aCharCollection isKindOf: CharacterCollection)
   ifTrue:[ ^ aCharCollection = self]
   ifFalse:[ ^ false]
%

category: 'Accessing'
method: MultiByteString
numArgs

"Returns the number of arguments the receiver would take, if the receiver were
 a message selector."

self isInfix ifTrue: [ ^1 ].
self isKeyword ifTrue: [ ^self occurrencesOf: $: ].
^0
%

category: 'Searching'
method: MultiByteString
speciesForCollect

"(R) Returns a class, an instance of which should be used as the result
 of collect: or other projections applied to the receiver."

^ self subclassResponsibility: #speciesForCollect
%

category: 'Adding'
method: MultiByteString
add: aCharOrCharColl

"(R) Appends all of the elements of aCharOrCharColl to the receiver and returns
 aCharOrCharColl."

<primitive: 437> 

aCharOrCharColl _validateClass: CharacterCollection .
^ self add: (aCharOrCharColl asMultiByteString: self)
%

category: 'Adding'
method: MultiByteString
addAll: aCharOrCharColl

"Equivalent to add: aCharOrCharColl."

<primitive: 437> 

(aCharOrCharColl isKindOf: QuadByteString) ifTrue:[
  ^ self _convertToQuadByte addAll: aCharOrCharColl ].

aCharOrCharColl _validateClass: CharacterCollection .
^ self add: (aCharOrCharColl asMultiByteString: self)
%

category: 'Adding'
method: MultiByteString
addLast: aCharOrCharColl

"Equivalent to add: aCharOrCharColl."

<primitive: 437> 

aCharOrCharColl _validateClass: CharacterCollection .
^ self add: (aCharOrCharColl asMultiByteString: self)
%

category: 'Accessing'
method: MultiByteString
size

"Returns the number of double-byte Characters in the receiver, which is also
 half the number of bytes that the receiver occupies."

^ self _basicSize // (self charSize)
%

category: 'Updating'
method: MultiByteString
size: anInteger

"Changes the size of the receiver to anInteger.

 If anInteger is less than the current size of the receiver, the receiver is
 shrunk accordingly.  If anInteger is greater than the current size of the
 receiver, the receiver is extended and new elements are initialized to nil."

^ self subclassResponsibility: #size:
%

category: 'Accessing'
method: MultiByteString
valueAt: anIndex

"Returns the value of the (multi-byte) Character at anIndex."

^ (self at: anIndex) asciiValue
%

category: 'Copying'
method: MultiByteString
withLFs

"Returns a copy of the receiver with all back-slashes replaced by line-feeds."

^self copyReplaceAll: '\' with: (self class with: Character lf)
%

category: 'Private'
method: MultiByteString
_at: anIndex equals: aCharCollection ignoreCase: aBoolean

"Returns true if aCharCollection is contained in the receiver, starting at
 anIndex.  Returns false otherwise.  The comparison is case-insensitive."

<primitive: 438>

anIndex _validateClass: Integer.
((anIndex <= 0) _or: [(anIndex > self size)])
ifTrue: [ ^ self _errorIndexOutOfRange: anIndex ].
aBoolean ifTrue:[ ^ self _primitiveFailed: #_at:equals:ignoreCase: ]
         ifFalse:[ ^ super at: anIndex equals: aCharCollection ].
%

category: 'Case-Insensitive Comparisons'
method: MultiByteString
at: anIndex equalsNoCase: aCharCollection

"Returns true if aCharCollection is contained in the receiver, starting at
 anIndex.  Returns false otherwise.  The comparison is case-insensitive."

^ self _at: anIndex equals: aCharCollection ignoreCase: true
%

category: 'Case-Sensitive Comparisons'
method: MultiByteString
at: anIndex equals: aCharCollection

"Returns true if aCharCollection is contained in the receiver, starting at
 anIndex.  Returns false otherwise.  The comparison is case-sensitive."

^ self _at: anIndex equals: aCharCollection ignoreCase: false
%

category: 'Case-Insensitive Comparisons'
method: MultiByteString
< aCharCollection

"Returns true if the receiver collates before the argument.
 Returns false otherwise.

 The comparison is case-insensitive."

"The ASCII letters A and a collate the same.
   'A' < 'a' returns false.
   'A' < 'b' returns true. "

<primitive: 440>

aCharCollection _validateClass: CharacterCollection .
^ aCharCollection > self
%

category: 'Case-Insensitive Comparisons'
method: MultiByteString
<= aCharCollection

"Returns true if the receiver collates before the argument or if all of the
 corresponding Characters in the receiver and argument are equal.
 Returns false otherwise.

 The comparison is consistent with that defined for the < method."

^(self > aCharCollection) not
%

category: 'Other Comparisons'
method: MultiByteString
collatingTable

"Return the appropriate collating table for this class"

^ self subclassResponsibility: #collatingTable
%

category: 'Other Comparisons'
method: MultiByteString
asciiLessThan: aString

"Returns true if the receiver collates before the argument using the
 ASCII collating table, which collates AB...Z...ab..z."

^ self lessThan: aString collatingTable: (self collatingTable)
%

category: 'Case-Insensitive Comparisons'
method: MultiByteString
> aCharCollection

"Returns true if the receiver collates after the argument.  Returns false
 otherwise.

 The comparison is case-insensitive."

<primitive: 442>

aCharCollection _validateClass: CharacterCollection.
^ aCharCollection < self
%

category: 'Case-Insensitive Comparisons'
method: MultiByteString
>= aCharCollection

"Returns true if the receiver collates after the argument or if all of the
 corresponding Characters in the receiver and argument are equal.
 Returns false otherwise.

 The comparison is consistent with that defined for the < method."

^ (self < aCharCollection) not
%

category: 'Other Comparisons'
method: MultiByteString
lessThan: aString collatingTable: aTable

"Returns true if the receiver collates before the argument.

 The aString argument may be a String or a MultiByteString.

 The collating sequence is defined by aTable, which must be an instance of
 MultiByteString.  The argument aTable is accessed with wordAt: to retrieve the
 collating sequence for a particular Character.  Use wordAt:put: to populate it.
 Characters beyond the range of aTable are collated using their raw character
 value."

<primitive: 444 >

aTable _validateClass: MultiByteString .
aString _validateClass: String .
self _primitiveFailed: #lessThan:collatingTable:
%

category: 'Other Comparisons'
method: MultiByteString
greaterThan: aString collatingTable: aTable

"Returns true if the receiver collates after the argument.

 The aString argument may be a String or a MultiByteString.

 The collating sequence is defined by aTable, which must be an instance of
 MultiByteString.  The argument aTable is accessed with wordAt: to retrieve the
 collating sequence for a particular Character.  Use wordAt:put: to populate it.
 Characters beyond the range of aTable are collated using their raw character
 value."

<primitive: 445 >

aTable _validateClass: MultiByteString .
aString _validateClass: String .
self _primitiveFailed: #greaterThan:collatingTable:
%

category: 'Other Comparisons'
method: MultiByteString
greaterThanOrEqual: aString collatingTable: aTable

"Returns true if the receiver collates after or the same as the argument.

 The aString argument may be a String or a MultiByteString.

 The collating sequence is defined by aTable, which must be an instance of
 MultiByteString.  The argument aTable is accessed with wordAt: to retrieve the
 collating sequence for a particular Character.  Use wordAt:put: to populate it.
 Characters beyond the range of aTable are collated using their raw character
 value."

^ (self lessThan: aString collatingTable: aTable) not
%

category: 'Other Comparisons'
method: MultiByteString
lessThanOrEqual: aString collatingTable: aTable

"Returns true if the receiver collates before or the same as the argument.

 The aString argument may be a String or a MultiByteString.

 The collating sequence is defined by aTable, which must be an instance of
 MultiByteString.  The argument aTable is accessed with wordAt: to retrieve the
 collating sequence for a particular Character.  Use wordAt:put: to populate it.
 Characters beyond the range of aTable are collated using their raw character
 value."

^ (self greaterThan: aString collatingTable: aTable) not
%

category: 'Other Comparisons'
method: MultiByteString
equals: aString collatingTable: aTable

"Returns true if the receiver collates the same as the argument.

 The aString argument may be a String or a MultiByteString.

 The collating sequence is defined by aTable, which must be an instance 
 of Double/QuadByteString.  The argument aTable is accessed with wordAt: 
 to retrieve the collating sequence for a particular Character.  Use 
 wordAt:put: to populate it.  Characters beyond the range of aTable 
 are collated using their raw character value."

<primitive: 447 >

aTable _validateClass: MultiByteString .
aString _validateClass: String .
self _primitiveFailed: #greaterThan:collatingTable:
%

category: 'Copying'
method: MultiByteString
copyFrom: index1 to: index2 into: anObject startingAt: index3

"Copies the elements of the receiver between index1 and index2, inclusive, into
 anObject starting at index3, overwriting the previous contents.  If anObject is
 the same object as the receiver, the source and destination blocks may
 overlap.

 The argument anObject should be a SequenceableCollection."

<primitive: 439>

(index1 > index2)
  ifTrue: [ ^ index1 _error: #rtErrBadCopyFromTo args: #[index2]].

(index1 < 1)
  ifTrue: [ ^ self _errorIndexOutOfRange: index1].

(index2 > self size)
  ifTrue: [ ^ self _errorIndexOutOfRange: index2].

((index3 < 1) _or: [(index3 > (anObject size + 1))])
  ifTrue: [ ^ anObject _errorIndexOutOfRange: index3].

^ super copyFrom: index1
              to: index2
            into: anObject
      startingAt: index3.
%

! deleted _idxCompareLessThan: v2.0

!deleted _idxCompareGreaterThan: v2.0

category: 'Case-Insensitive Comparisons'
method: MultiByteString
equalsNoCase: aCharCollection

"Returns true if corresponding Characters in the receiver and argument are equal
 and aCharCollection is comparable with the receiver, and aCharCollection is
 not a kind of Symbol.  Returns false otherwise.

 The comparison for equal is case-insensitive.

 Note that 'kind of Symbol' means either an instance of Symbol or instance of
 DoubleByteSymbol."

<primitive: 450>

(aCharCollection isKindOf: String) ifTrue:[
  self _primitiveFailed: #equalsNoCase: .
  ^ self _uncontinuableError
  ].
^super isEquivalent: aCharCollection
%

category: 'Case-Insensitive Comparisons'
method: MultiByteString
isEquivalent: aCharCollection
    "Returns true if the receiver is equivalent to aCharCollection.
    The receiver is equivalent to aCharCollection if the receiver
    contains the same Characters as aCharCollection regardless of case
    or internal representation.  For example, if $a is in
    aCharCollection, it is equivalent to any representation of an 'a'
    in the receiver's character set."

    ^self equalsNoCase: aCharCollection
%

category: 'Private'
method: MultiByteString
_findSmallString: subString startingAt: startIndex ignoreCase: aBoolean

"If a receiver contains subString beginning at some point at or after
 startIndex, this returns the index at which subString begins.  If the
 receiver does not contain subString, this returns 0."

<primitive: 451>

startIndex _validateClass: Integer.
(startIndex < 1) | (startIndex > self size)
ifTrue: [ ^ self _error: #objErrBadOffsetIncomplete args: #[startIndex] ].
subString _validateClasses:[ String, MultiByteString ].
subString basicSize > 16272  "virtual machine constant, OBJ_BYTES_SIZE" ifTrue:[
  subString _halt: 'String too large for string search primitive.'
  ].
aBoolean _validateClass: Boolean .
^ self _primitiveFailed: #_findSmallString:startingAt:
%

category: 'Case-Sensitive Searching'
method: MultiByteString
includes: aCharacter

"Returns true if the receiver contains aCharacter, false otherwise.
 The search is case-sensitive."

^(aCharacter isKindOf: AbstractCharacter) and:
	[self includesValue: aCharacter].
%

category: 'Case-Sensitive Searching'
method: MultiByteString
includesValue: aCharacter

"Returns true if the receiver contains aCharacter.  The search is
 case-sensitive."

<primitive: 452>

"the following code is executed if aCharacter is a JISCharacter "

aCharacter _validateClass: AbstractCharacter .
^ self includesValue: aCharacter asCharacter .
%

category: 'Case-Sensitive Searching'
method: MultiByteString
indexOf: aCharacter startingAt: startIndex

"Returns the index of the first occurrence of aCharacter in the receiver,
 not preceding startIndex.  If the receiver does not contain aCharacter,
 returns zero.

 The search is case sensitive."

<primitive: 453>
startIndex _validateClass: Integer .
startIndex < 1 ifTrue:[ ^ self _errorIndexOutOfRange: startIndex].
^ self indexOf: aCharacter asCharacter startingAt: startIndex
%

category: 'Adding'
method: MultiByteString
insertAll: aCharOrCharColl at: anIndex

"Inserts aCharOrCharColl at the specified index.  Returns aCharOrCharColl."

<primitive: 454>

anIndex _validateClass: Integer .
((anIndex <= 0) _or: [anIndex > (self size + 1)])
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].

(aCharOrCharColl isKindOf: QuadByteString) ifTrue:[
  ^ self _convertToQuadByte insertAll: aCharOrCharColl at: anIndex ].

aCharOrCharColl _validateClasses: #[AbstractCharacter, CharacterCollection].
^self insertAll: (aCharOrCharColl asMultiByteString: self)  at: anIndex.
%

category: 'Converting'
method: MultiByteString
asUppercase

"Returns a new instance of the receiver's class, with all lower-case Characters
 in the receiver changed to upper-case.  Lower-case Characters are the
 ASCII characters $a to $z inclusive.  If the receiver is a MultiByteSymbol, 
 returns an instance of MultiByteString."

<primitive: 455>

self _primitiveFailed: #asUppercase .
self _uncontinuableError
%

! finish fixing 9288
category: 'Execution'
method: MultiByteString
_compileInContext: anObject symbolList: aSymbolList

"Returns a GsMethod that is not in the method dictionary of any class, or
 else generates an error if there are compilation errors."

<primitive: 150>

aSymbolList _validateClass: SymbolList .
^ self _primitiveFailed: #_compileInContext:symbolList:
%

category: 'Execution'
method: MultiByteString
evaluate

"Compiles the receiver as an unbound method and executes it using the current
 default symbol list."

^ (self _compileInContext: nil 
       symbolList: GsSession currentSession symbolList)
  _executeInContext: nil
%

category: 'Execution'
method: MultiByteString
evaluateInContext: anObject symbolList: aSymbolList

"Compiles the receiver as an instance method for the class of anObject, using
 aSymbolList as the symbol list.  Executes the resulting GsMethod using anObject
 as self and returns the result of the execution.  Generates an error if
 compilation errors occur."

^ (self _compileInContext: anObject symbolList: aSymbolList)
    _executeInContext: anObject
%

category: 'Hashing'
method: MultiByteString
hash

"Returns a positive Integer based on a case-sensitive hash of the contents 
 of the receiver."

"The algorithm implemented is described in:

 [Pearson 90]
 Pearson, Peter K., Fast Hashing of Variable-Length Text Strings,
 Communications of the ACM 33, 6, (June 1990), 677-680."

<primitive: 282>
self _primitiveFailed: #hash .
self _uncontinuableError
%

category: 'Storing and Loading'
classmethod: MultiByteString
loadFrom: passiveObj

"Reads from passiveObj the passive form of an object.  Converts the object to
 its active form by loading the information into a new instance of the receiver.
 Returns the new instance."

| inst |

inst := String new: passiveObj readSize . "yes, make a String !!"
inst loadFrom: passiveObj.
inst _changeClassToMultiByte: self .  "Now change to double/quad byte"
^inst
%

category: 'Storing and Loading'
method: MultiByteString
writeTo: passiveObj

"Converts the receiver to its passive form and writes that information on
 passiveObj."

passiveObj writeClass: self class;
  writeSize: self _basicSize;
    nextPutAllBytes: self; space
%
category: 'Removing'
method: MultiByteString
removeFrom: startIndex to: stopIndex

"(R) Deletes the elements of the receiver from startIndex to stopIndex.  The
 size of the receiver is decreased by stopIndex - startIndex + 1.

 Both startIndex and stopIndex must be positive integers not larger than the
 size of the receiver, with startIndex <= stopIndex."

<primitive: 271>

(stopIndex < startIndex)
ifTrue:
   [ ^ startIndex _error: #rtErrBadCopyFromTo args: #[stopIndex]].
((stopIndex > self size) _or: [(stopIndex < 1)])
   ifTrue: [ ^ self _errorIndexOutOfRange: stopIndex].
(startIndex < 1)
   ifTrue: [ ^ self _errorIndexOutOfRange: startIndex].
^ self _primitiveFailed: #removeFrom:to:
%
category: 'Indexing Support'
method: MultiByteString
_changingSizeTo: newSize
 
"Notifies any modification tracking objects that the receiver (an indexable
 object) is having its size changed."
 
"the object manager reports size changes in number of bytes; convert this
 to number of characters."

^ super _changingSizeTo: newSize // (self charSize)
%
category: 'Testing'
method: MultiByteString
containsSeparator

"Returns true if the receiver contains a separator Character."

1 to: self size do: [:i |
  (self at: i) isSeparator ifTrue: [^true].
].
^false
%

category: 'New Indexing Comparison - prims'
method: MultiByteString
_idxPrimCompareLessThan: aCharCollection

"This comparison operation is used for the indexing subsystem to
 determine an ordering for insertion into indexing objects.

 This method collates letters AaBb..Zz."

"The comparison should be compatible with the case-insensitive semantics
 of the String method with selector #< .
 Same primitive as String>>lessThan: "

<primitive: 441>

aCharCollection == nil ifTrue: [ ^ false ].
aCharCollection _validateClass: CharacterCollection.
^ aCharCollection > self
%

category: 'New Indexing Comparison - prims'
method: MultiByteString
_idxPrimCompareGreaterThan: aCharCollection

"This comparison operation is used for the indexing subsystem to
 determine an ordering for insertion into indexing objects such as B-trees.

 This method collates letters AaBb..Zz."

"The comparison should be compatible with the case-insensitive semantics
 of the String method with selector #> .
 Same primitive as String>>greaterThan: ."

<primitive: 443>
aCharCollection == nil ifTrue: [ ^ true ].
aCharCollection _validateClass: CharacterCollection.
^ aCharCollection < self
%

category: 'Encoding'
method: MultiByteString
encodeAsUTF8

"Encode receiver in UTF8 format. If all characters in receiver are US-ASCII,
 answer the receiver. Result is a String." 

^self _unicodePrim: 0
%

category: 'Encoding'
method: MultiByteString
_unicodePrim: opCode

"opCode 0 - Encode receiver in UTF8 format. If all characters in receiver are US-ASCII,
            answer the receiver. Result is a String.
 opCode 1 - Not a valid opCode for MultiByteString. See String>>unicodePrim:
 opCode 2 - Not a valid opCode for MultiByteString. See String>>unicodePrim:" 

<primitive: 468>
self _primitiveFailed: #_unicodePrim:
%
category: 'Encoding'
method: MultiByteString
encodeUsing: encodingArray

  "A new MultiByteString will be returned where each character in the receiver is
   replaced by the character or string from encodingArray that is located
   at the position in encodingArray for the character interpreted as a byte.
   If the byte value is greater than the table size, nil is returned as the
   result."

  <primitive: 654>
  | stream charOrString |
  stream := WriteStream on: MultiByteString new.
  self do:
    [:char |
      charOrString := encodingArray at: char asInteger + 1 ifAbsent: [ ^nil].
      charOrString _class == Character
        ifTrue: [stream nextPut: charOrString]
        ifFalse: [stream nextPutAll: charOrString].
    ].
  ^stream contents
%

